-- simple wave file library
-- by Daryl Van Den Brink
-- vandenbrink@micronet.net.au
-- or EUforum@topica.com if that address doesn't
-- exist anymore (I'm thinking of changing ISP)

include get.e
include machine.e
include wildcard.e

function readTag(integer fn)
    return get_bytes(fn, 4)
end function

function readInt(integer fn)
    sequence bytes
    bytes = get_bytes(fn, 4)
    if length(bytes) < 4 then
        return -1
    end if
    return bytes_to_int(bytes)
end function

function readShort(integer fn)
    return getc(fn) + getc(fn)*256
end function

procedure writeShort(integer fn, integer i)
    integer low
    low = and_bits(i, #FF)
    puts(fn, {low, (i-low) / 256})
end procedure

global function read_wave_file(sequence filename)
-- read WAV file
-- if successful return a sequence of the following format:
-- {channels, sample_rate, bits, length, wave_data}
-- where wave_data is of the form {channel1, channel2, ...}
-- If the read failed, return an empty sequence
--
-- structure of wave file header record
-- tagged record type file. tags are 4 characters long.
--     data              offset       data nature
--     ------------------------------------------
--     tag               [1..4]       "RIFF" file Id tag
--     integer           [5..8]       file length excluding this record
--     tag               [9..12]      "WAVE"  confirm it's a wave file
--     tag               [13..16]     "fmt " announce format record
--     integer           [17..20]     length of format record
--     word              [21..22]     format tag
--     word              [23..24]     number of channels  (mono=1,stereo=2)
--     integer           [25..28]     samples per seconds
--     integer           [29..32]     average bytes per second  
--     word              [33..34]     block align
--     word              [35..36]     bits per sample
--     tag               [37..40]     "data " or "fact "  record
--     integer           [41..44]     length of that record, if data it means
--                                    length of sound sample, if fact jump
--                                     over it and go to data record.   
--
    object tmp
    integer fmtlen, datalen, block_align, sample
    integer channels, sps, bits, wavelen
    sequence wave
    integer fn
    
    fn = open(filename, "rb")
    if fn = -1 then
        return {}
    end if
    
    tmp = readTag(fn) --"RIFF"
    if compare(tmp, "RIFF") then
        close(fn)
        return {}
    end if
    tmp = readInt(fn) --length of file - 8
    tmp = readTag(fn) --"WAVE"
    if compare(tmp, "WAVE") then
        close(fn)
        return {}
    end if
    tmp = readTag(fn) --"fmt "
    if compare(tmp, "fmt ") then
        close(fn)
        return {}
    end if
    tmp = readInt(fn) --length of format record
    if tmp < 16 then
        close(fn)
        return {}
    end if
    fmtlen = tmp
    tmp = readShort(fn) --format tag (1 = PCM)
    if tmp != 1 then
        close(fn)
        return {}
    end if
    tmp = readShort(fn) --number of channels
    if tmp <= 0 or tmp > 2 then
        close(fn)
        return {}
    else
        channels = tmp
    end if
    sps = readInt(fn) --samples per second
    if sps <= 0 then
        close(fn)
        return {}
    end if
    tmp = readInt(fn) --average bytes per second
    block_align = readShort(fn) --block align
    bits = readShort(fn) --bits per sample
    if not find(bits, {8, 16, 24}) then
        close(fn)
        return {}
    end if
    if fmtlen > 16 then -- bypass rest of format record
        for n = 17 to fmtlen do
            tmp = getc(fn)
        end for
    end if
    tmp = readTag(fn) --"data" or "fact"
    while compare(lower(tmp), "data") do
        if length(tmp) < 4 then
            close(fn)
            return {}
        end if
        tmp = readInt(fn)
        if tmp < 0 then
            close(fn)
            return {}
        end if
        for n = 1 to tmp do
            tmp = getc(fn)
        end for
        if tmp = {} then
            close(fn)
            return {}
        end if
        tmp = readTag(fn)
    end while
    datalen = readInt(fn)
    datalen = datalen - remainder(datalen, block_align)
    wave = repeat(0, datalen)
    
    if block_align != channels * bits / 8 then
        close(fn)
        return {}
    end if
    
    if bits = 8 then
        wavelen = floor(datalen / channels)
        wave = repeat(repeat(0, wavelen), channels)
        for n = 1 to wavelen do
            for m = 1 to channels do
                wave[m][n] = getc(fn) - 128
            end for
        end for
    elsif bits = 16 then
        wavelen = floor(datalen / 2 / channels)
        wave = repeat(repeat(0, wavelen), channels)
        for n = 1 to wavelen do
            for m = 1 to channels do
                sample = getc(fn) + getc(fn) * 256
                if sample > 32767 then
                    wave[m][n] = sample - 65536
                else
                    wave[m][n] = sample
                end if
            end for
        end for
    elsif bits = 24 then
        wavelen = floor(datalen / 3 / channels)
        wave = repeat(repeat(0, wavelen), channels)
        for n = 1 to wavelen do
            for m = 1 to channels do
                sample = getc(fn) + getc(fn) * 256 + getc(fn) * 65536
                if sample > 8388607 then
                    wave[m][n] = sample - 16777216
                else
                    wave[m][n] = sample
                end if
            end for
        end for
    end if
    
    close(fn)
    return {channels, sps, bits, wavelen, wave}
end function -- read_wave_file

global function write_wave_file(sequence filename, sequence wavedata)
    -- write wave data to a wave file
    -- return 0 if successful, -1 if the file couldn't be created
    integer datasize, filesize, block_align, x
    integer channels, sps, bits, wavelen
    integer fn
    
    fn = open(filename, "wb")
    if fn = -1 then
        return -1
    end if
    
    channels = wavedata[1]
    sps = wavedata[2]
    bits = wavedata[3]
    wavelen = wavedata[4]
    
    block_align = channels * bits / 8
    datasize = wavelen * block_align
    filesize = datasize + 44
    
    puts(fn, "RIFF" & int_to_bytes(filesize - 8) & "WAVE")
    puts(fn, "fmt " & int_to_bytes(16))
    writeShort(fn, 1)        -- format tag
    writeShort(fn, channels) -- number of channels
    puts(fn, int_to_bytes(sps))  -- samples per second
    puts(fn, int_to_bytes(sps * block_align))    -- average bytes per second
    writeShort(fn, block_align) -- block align
    writeShort(fn, bits)        -- bits per sample
    puts(fn, "data" & int_to_bytes(datasize))
    
    if bits = 8 then
        for n = 1 to wavelen do
            for m = 1 to channels do
                puts(fn, wavedata[5][m][n] + 128)
            end for
        end for
    elsif bits = 16 then
        for n = 1 to wavelen do
            for m = 1 to channels do
                x = wavedata[5][m][n]
                if x < 0 then
                    x += 65536
                end if
                puts(fn, {and_bits(x, #FF), floor(x / 256)})
            end for
        end for
    elsif bits = 24 then
        for n = 1 to wavelen do
            for m = 1 to channels do
                x = wavedata[5][m][n]
                if x < 0 then
                    x += 16777216
                end if
                puts(fn, {and_bits(x, #FF), floor(and_bits(x, #FF00) / 256), floor(x/65536)})
            end for
        end for
    end if
    
    close(fn)
    return 0
end function -- write_wave_file

